#include "DirectXMeshPainter.h"

#include "DrawingContext.h"
#include "DrawingSurface.h"
#include "SceneModel.h"

#include "TriMesh.h"
#include "LineMesh.h"

#include "DirectXShaderProgram.h"

#include <windows.h>
#include <iostream>
#include <filesystem>

#include <directx/d3dx12.h>

DirectXMeshPainter::DirectXMeshPainter()
{
    const float aspectRatio = 1.0;
    mVertexBuffer = 
    {
        { { 0.0f, 0.25f * aspectRatio, 0.0f }, { 1.0f, 0.0f, 0.0f, 1.0f } },
        { { 0.25f, -0.25f * aspectRatio, 0.0f }, { 0.0f, 1.0f, 0.0f, 1.0f } },
        { { -0.25f, -0.25f * aspectRatio, 0.0f }, { 0.0f, 0.0f, 1.0f, 1.0f } }
    };

    auto shader_path = std::filesystem::path(__FILE__).parent_path() / "shaders.hlsl";
    mShaderProgram = std::make_unique<DirectXShaderProgram>(shader_path, shader_path);
}

void DirectXMeshPainter::initializeShader()
{

}

DirectXShaderProgram* DirectXMeshPainter::getShaderProgram() const
{
    return mShaderProgram.get();
}

unsigned int DirectXMeshPainter::getVertexBufferSize() const
{
    return mVertexBuffer.size()*sizeof(Vertex);
}

unsigned int DirectXMeshPainter::getVertexSize() const
{
    return sizeof(Vertex);
}

ID3D12PipelineState* DirectXMeshPainter::getPipelineState() const
{
    return mPipelineState.Get();
}

ID3D12RootSignature* DirectXMeshPainter::getRootSignature() const
{
    return mRootSignature.Get();
}

void DirectXMeshPainter::updateVertexBuffer(unsigned char* pBuffer) const
{
    memcpy(pBuffer, mVertexBuffer.data(), getVertexBufferSize());
}

void DirectXMeshPainter::updateCommandList(const CD3DX12_CPU_DESCRIPTOR_HANDLE& rtvHandle, ID3D12GraphicsCommandList* commandList)
{
    const float clearColor[] = { 0.0f, 0.2f, 0.4f, 1.0f };
    commandList->ClearRenderTargetView(rtvHandle, clearColor, 0, nullptr);
    commandList->IASetPrimitiveTopology(D3D_PRIMITIVE_TOPOLOGY_TRIANGLELIST);
    commandList->IASetVertexBuffers(0, 1, &mVertexBufferView);
    commandList->DrawInstanced(3, 1, 0, 0);
}

void DirectXMeshPainter::initializeBuffers()
{
    /*
    glGenBuffers(1, &mVertexBuffer);

    glGenBuffers(1, &mElementBuffer);

    glGenVertexArrays(1, &mVertexArray);
    */
}

void DirectXMeshPainter::paint(const std::vector<float>& verts, const std::vector<unsigned>& elements, const std::vector<float>& color, bool lines)
{
    /*
    glBindVertexArray(mVertexArray);

    glBindBuffer(GL_ARRAY_BUFFER, mVertexBuffer);
    glBufferData(GL_ARRAY_BUFFER, verts.size() * sizeof(float), verts.data(), GL_STATIC_DRAW);

    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, mElementBuffer);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, elements.size() * sizeof(unsigned), elements.data(), GL_STATIC_DRAW);

    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 3 * sizeof(float), (void*)0);
    glEnableVertexAttribArray(0);

    //glm::mat4 projection = glm::ortho(0.0f, width, 0.0f, height);
    //glUniformMatrix4fv(glGetUniformLocation(mShaderProgram->getHandle(), "projection"), 1, GL_FALSE, glm::value_ptr(projection));

    glUseProgram(mShaderProgram->getHandle());
    glBindVertexArray(mVertexArray);

    int vertexColorLocation = glGetUniformLocation(mShaderProgram->getHandle(), "ourColor");
    glUniform4f(vertexColorLocation, float(color[0]), float(color[1]), float(color[2]), float(color[3]));

    //glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

    if (lines)
    {
        glDrawElements(GL_LINES, elements.size(), GL_UNSIGNED_INT, 0);
    }
    else
    {
        glDrawElements(GL_TRIANGLES, elements.size(), GL_UNSIGNED_INT, 0);
    }
    glBindVertexArray(0);
    */
}

void DirectXMeshPainter::paint(SceneModel* model, DrawingContext* context)
{
    return;
    /*
    if (!mShaderProgram)
    {
        initializeShader();
    }

    if (mVertexArray == 0)
    {
        initializeBuffers();
    }

    auto surface = context->getSurface();
    const auto width = float(surface->getWidth());
    const auto height = float(surface->getHeight());

    auto transform = model->getTransform();
    auto vertices = model->getMesh()->getVerticesFlat<float>();
    for (std::size_t idx = 0; idx < vertices.size(); idx++)
    {
        if (idx % 3 == 0)
        {
            auto x = vertices[idx] * transform.getScaleX() + transform.getLocation().getX();
            vertices[idx] = 2 * x / width - 1.0;
        }
        else if (idx % 3 == 1)
        {
            auto y = vertices[idx] * transform.getScaleY() + transform.getLocation().getY();
            vertices[idx] = 1.0 - 2 * y / height;
        }
    }

    std::vector<unsigned> indices;
    const bool line_mesh = model->getMesh()->getType() == AbstractMesh::MeshType::LINE;
    if (line_mesh)
    {
        indices = dynamic_cast<LineMesh*>(model->getMesh())->getEdgeNodeIds();
    }
    else
    {
        indices = dynamic_cast<TriMesh*>(model->getMesh())->getFaceNodeIds();
    }

    auto model_color = model->getColor().getAsVectorDouble();
    std::vector<float> color = { float(model_color[0]), float(model_color[1]), float(model_color[2]), float(model_color[3]) };

    paint(vertices, indices, color, line_mesh);

    if (model->getShowOutline())
    {
        auto edge_indices = dynamic_cast<TriMesh*>(model->getMesh())->getEdgeNodeIds();
        paint(vertices, edge_indices, { 0, 0, 0, 1 }, true);
    }
    */
}

void DirectXMeshPainter::createVertexBuffer(ID3D12Device* device)
{
    // Note: using upload heaps to transfer static data like vert buffers is not 
    // recommended. Every time the GPU needs it, the upload heap will be marshalled 
    // over. Please read up on Default Heap usage. An upload heap is used here for 
    // code simplicity and because there are very few verts to actually transfer.
    CD3DX12_HEAP_PROPERTIES heapProps(D3D12_HEAP_TYPE_UPLOAD);
    auto desc = CD3DX12_RESOURCE_DESC::Buffer(getVertexBufferSize());
    device->CreateCommittedResource(&heapProps, D3D12_HEAP_FLAG_NONE, &desc, D3D12_RESOURCE_STATE_GENERIC_READ, nullptr, IID_PPV_ARGS(&mD3dVertexBuffer));

    // Copy the mesh data to the vertex buffer.
    UINT8* pVertexDataBegin;
    CD3DX12_RANGE readRange(0, 0);        // We do not intend to read from this resource on the CPU.
    mD3dVertexBuffer->Map(0, &readRange, reinterpret_cast<void**>(&pVertexDataBegin));
    updateVertexBuffer(pVertexDataBegin);
    mD3dVertexBuffer->Unmap(0, nullptr);

    // Initialize the vertex buffer view.
    mVertexBufferView.BufferLocation = mD3dVertexBuffer->GetGPUVirtualAddress();
    mVertexBufferView.StrideInBytes = getVertexSize();
    mVertexBufferView.SizeInBytes = getVertexBufferSize();
}

void DirectXMeshPainter::createRootSignature(ID3D12Device* device)
{
    CD3DX12_ROOT_SIGNATURE_DESC rootSignatureDesc;
    rootSignatureDesc.Init(0, nullptr, 0, nullptr, D3D12_ROOT_SIGNATURE_FLAG_ALLOW_INPUT_ASSEMBLER_INPUT_LAYOUT);

    Microsoft::WRL::ComPtr<ID3DBlob> signature;
    Microsoft::WRL::ComPtr<ID3DBlob> error;
    D3D12SerializeRootSignature(&rootSignatureDesc, D3D_ROOT_SIGNATURE_VERSION_1, &signature, &error);
    device->CreateRootSignature(0, signature->GetBufferPointer(), signature->GetBufferSize(), IID_PPV_ARGS(&mRootSignature));
}

void DirectXMeshPainter::createPipelineStateObject(ID3D12Device* device)
{
    auto vert_shader = mShaderProgram->getVertexShader();
    auto pixel_shader = mShaderProgram->getPixelShader();

    D3D12_INPUT_ELEMENT_DESC inputElementDescs[] =
    {
        { "POSITION", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 0, D3D12_INPUT_CLASSIFICATION_PER_VERTEX_DATA, 0 },
        { "COLOR", 0, DXGI_FORMAT_R32G32B32A32_FLOAT, 0, 12, D3D12_INPUT_CLASSIFICATION_PER_VERTEX_DATA, 0 }
    };

    D3D12_GRAPHICS_PIPELINE_STATE_DESC psoDesc = {};
    psoDesc.InputLayout = { inputElementDescs, _countof(inputElementDescs) };
    psoDesc.pRootSignature = mRootSignature.Get();
    psoDesc.VS = CD3DX12_SHADER_BYTECODE(vert_shader);
    psoDesc.PS = CD3DX12_SHADER_BYTECODE(pixel_shader);
    psoDesc.RasterizerState = CD3DX12_RASTERIZER_DESC(D3D12_DEFAULT);
    psoDesc.BlendState = CD3DX12_BLEND_DESC(D3D12_DEFAULT);
    psoDesc.DepthStencilState.DepthEnable = FALSE;
    psoDesc.DepthStencilState.StencilEnable = FALSE;
    psoDesc.SampleMask = UINT_MAX;
    psoDesc.PrimitiveTopologyType = D3D12_PRIMITIVE_TOPOLOGY_TYPE_TRIANGLE;
    psoDesc.NumRenderTargets = 1;
    psoDesc.RTVFormats[0] = DXGI_FORMAT_R8G8B8A8_UNORM;
    psoDesc.SampleDesc.Count = 1;
    device->CreateGraphicsPipelineState(&psoDesc, IID_PPV_ARGS(&mPipelineState));
}
