#include "DirectXPainter.h"

#include "DrawingContext.h"
#include "DrawingSurface.h"
#include "Scene.h"
#include "SceneItem.h"
#include "SceneModel.h"
#include "SceneText.h"

#include "TriMesh.h"

#include "FontsManager.h"
#include "FontGlyph.h"
#include "TextData.h"

#include "DirectXShaderProgram.h"
#include "DirectXMeshPainter.h"
#include "DirectXTextPainter.h"

#include "File.h"

#include <windows.h>

#include <iostream>

DirectXPainter::DirectXPainter()
    :   mMeshPainter(std::make_unique<DirectXMeshPainter>()),
        mTextPainter(std::make_unique<DirectXTextPainter>())
{

}

DirectXMeshPainter* DirectXPainter::getMeshPainter() const
{
    return mMeshPainter.get();
}

DirectXTextPainter* DirectXPainter::getTextPainter() const
{
    return mTextPainter.get();
}

void DirectXPainter::paint(DrawingContext* context)
{
    auto surface = context->getSurface();
    const auto width = double(surface->getWidth());
    const auto height = double(surface->getHeight());

    /*
    glViewport(0, 0, width, height);
    glOrtho(0, width, 0, height, -1.0, 1.0);

    glClearColor(0.5, 0.5, 1.0, 0.0);
    glClear(GL_COLOR_BUFFER_BIT);
    */

    auto scene = context->getSurface()->getScene();
    for (const auto item : scene->getItems())
    {
        if (item->getType() == SceneItem::Type::MODEL)
        {
            mMeshPainter->paint(dynamic_cast<SceneModel*>(item), context);
        }
        else if (item->getType() == SceneItem::Type::TEXT)
        {
            mTextPainter->paint(dynamic_cast<SceneText*>(item), context);
        }
    }

    /*
    glFlush();
    */
}
